package com.drhoffmannstoolsdataloggerreader;

/* Logger.java (c) 2011-2020 by Markus Hoffmann
 *
 * This file is part of USB-Dataloggerreader for Android, (c) by Markus Hoffmann 2011-2020
 * ============================================================================
 * USB-Dataloggerreader for Android is free software and comes with 
 * NO WARRANTY - read the file COPYING for details.
 */ 

import java.nio.ByteBuffer;
import java.util.Calendar;
import java.util.Locale;

import android.hardware.usb.UsbConstants;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbEndpoint;
import android.hardware.usb.UsbInterface;
import android.hardware.usb.UsbManager;
import android.hardware.usb.UsbRequest;
import android.util.Log;


/* Diese Klasse beschreibt alles, was zu einem Logger bzw. dessen USB-Verbindung gehört */
public class Logger {
	private static final String TAG = "LOGGER";
	private static final int DEBUG=0;
	public USBDataloggerreaderActivity mActivity=null;


	public UsbManager UsbManager;
	public UsbDevice Device;
	private UsbDeviceConnection Connection=null;
	public String Serial_id="00_0000000000";
	public String Manufacturer="UKN";
	public String Product="unknown logger";
	public int Vid=-1;   /*Vendor ID*/
	public int Pid=-1;   /*Product ID*/
	private UsbEndpoint EndpointOut;
	private UsbEndpoint EndpointIn;
	public int packet_size=64;
	public int memory_size=0xfe00;

	/*Logger spezifisch*/
	public int loggertype=-1;
	public int protocol=0;
	public boolean do_repair=false;
	public boolean do_simulate=false;
	public boolean do_calibrate=false;
	public boolean be_quiet=false;
	public int simulate_type;

	/*Verschiedene Protokoll-Typen*/
	public final static int PROTO_VOLTCRAFT=0;
	public final static int PROTO_VOLTCRAFT_NEW=1;
	public final static int PROTO_VOLTCRAFT_WEATHER=2;
	public final static int PROTO_ELV=4;  /* Lascar logger*/
	public final static int PROTO_HID=5;  /* Freetec logger*/
	public final static int PROTO_TFD=6;  /* ELV TFD*/
	public final static int PROTO_CP210X=7;  /* e.g. Voltcraft DL 141*/

	/*Verschiedene Logger-Typen*/
	public final static int LTYP_TEMP=100;
	public final static int LTYP_TH=120;
	public final static int LTYP_THP=180;
	public final static int LTYP_VOLT=200;
	public final static int LTYP_CURR=300;
	public final static int LTYP_GAS=400;
	public final static int LTYP_SOUND=500;

	public boolean isconnected=false;
	public boolean isreadconfig=false;
	public boolean isreaddata=false;
	public boolean issaved=false;

	public final LoggerData data=new LoggerData();
	public final LoggerConfig config=new LoggerConfig();
	public final Simulator simulator=new Simulator(this);

	/*Calibrationsdaten dieses Loggers (darf von denen in den LoggerData differeieen)*/
	public float calibration_Avalue=(float)0.1; 
	public float calibration_Bvalue=(float)0.0; 
	public float calibration_A2value=(float)0.1; 
	public float calibration_B2value=(float)0.0; 

	public final static int WAIT_READ=500;
	public final static int WAIT_WRITE=100;

	/*Zum Debuggen: Kommunikationsprotokoll*/
	private String comproto="";

	public static String bytearray2string(byte[] a) {
		int i;
		String b="("+a.length+")";
		b=b+"{";
		for(i=0;i<a.length;i++) {
			b=b+String.format("0x%02x",a[i]);
			if(i+1<a.length) b=b+",";
		}
		b=b+"}";
		return b;
	}
	public String toString() {
		String mes="# Device Information:\n"+
				"vendor_id="+Vid+" ("+Manufacturer+")"+"\n"+
				"product_id="+Pid+" ("+Product+")"+"\n"+
				"serial_id="+Serial_id+"\n"+
				"protocol="+protocol+"\n"+
				"loggertype="+loggertype+"\n"+
				"memory_size="+memory_size+"\n"+
				"packet_size="+packet_size+"\n"+
				"version="+config.getversion()+"\n"+
				"cbuf=";
		if(config.configbuf==null) mes=mes+"null";
		else {
			mes=mes+bytearray2string(config.configbuf);
		}
		mes=mes+";\n"+
				"data_saved="+data.issaved+"\n"+
				"data_anzdata="+data.anzdata+"\n"+
				"data_quality="+data.quality+"\n"+
				"";
		if(comproto.length()>0) mes=mes+"\nCommunication protocol:\n"+comproto;
		return mes;
	}
	public void log2comproto(String a,int prio) {
		if(DEBUG>=prio) comproto=comproto+a+"\n";
		// if(DEBUG>prio) Log.d(TAG,a);
	}
	public boolean checkforUSBdevices() {
		boolean flag=false;
		log2comproto("Check for USB devices...",0);
		// check for existing devices
		for (UsbDevice device :  UsbManager.getDeviceList().values()) {
			flag|=filterdevice(device);
		}
		if(!flag && do_simulate) {
			/*Wenn noch eine alte Verbindung besteht, diese schliessen.*/
			if(Connection!=null) close();
			// setze simulation auf.
			simulator.init(simulate_type);
		}
		return(flag||do_simulate);
	}


	public boolean filterdevice(UsbDevice device) {
		log2comproto("Checking device: c:"+device.getDeviceClass()+
				", v:"+device.getVendorId()+", p:"+device.getProductId(),0);
		if(device.getVendorId()==4292 && device.getProductId()==2) {
			/*Die ELV Reihe*/
			protocol=PROTO_ELV;  /*ELV-USB*/
			setDevice(device);
			return true;
		} else if(device.getVendorId()==4292 && device.getProductId()==3) {
			protocol=PROTO_VOLTCRAFT;
			setDevice(device);
			return true;
		}	else if(device.getVendorId()==4292 && device.getProductId()==60000) {
			protocol=PROTO_CP210X;
			setDevice(device);
			return true;
		}	else if(device.getVendorId()==4292 && device.getProductId()==60001) {
			protocol=PROTO_VOLTCRAFT_NEW;
			setDevice(device);
			return true;
		}	else if(device.getVendorId()==6465 && (device.getProductId()==2010 || device.getProductId()==32801)) {
			protocol=PROTO_HID;
			setDevice(device);
			return true;
		}	else if(device.getVendorId()==1027 && device.getProductId()==23389) {
			protocol=PROTO_TFD;
			setDevice(device);
			return true;

		}
		return false;
	}


	public void setDevice(UsbDevice device) {
		/*Wenn noch eine alte Verbindung besteht, diese schliessen.*/
		if(Connection!=null) close();
		if(do_simulate) {
			simulator.init(simulate_type);
			return;
		}
		UsbEndpoint epOut = null;
		UsbEndpoint epIn = null;
		Vid=-1;
		Pid=-1;

		if(device==null) {
			mActivity.displaystatus(" - ",5);
			mActivity.displaymessage("ERROR: Device not connected.",5);
			return;
		}

		log2comproto("setDevice " + device,1);
		if (device.getInterfaceCount() != 1) {
			mActivity.displaymessage("ERROR: could not find interface",5);
			return;
		}

		UsbInterface intf = device.getInterface(0);
		if(protocol!=PROTO_HID) {
			// device should have two endpoints
			if (intf.getEndpointCount() != 2) {
				Log.e(TAG, "endpoints: "+intf.getEndpointCount());
				mActivity.displaymessage("ERROR: could not find endpoints",5);
				return;
			}
			// look for our bulk endpoints
			for (int i = 0; i < intf.getEndpointCount(); i++) {
				log2comproto("Endpoint: "+intf.toString(),2);
				UsbEndpoint ep = intf.getEndpoint(i);
				if (ep.getType() == UsbConstants.USB_ENDPOINT_XFER_BULK) {
					if (ep.getDirection() == UsbConstants.USB_DIR_OUT) {
						epOut = ep;
					} else {
						epIn = ep;
					}
				}
			}

			if (epOut == null || epIn == null) {
				throw new IllegalArgumentException("not all endpoints found.");
			} else log2comproto("Endpoints found.",1);


		} else {
			// device should have only one endpoints
			if (intf.getEndpointCount() != 1) {
				Log.e(TAG, "endpoint: "+intf.getEndpointCount());
				mActivity.displaymessage("ERROR: could not find endpoint",5);
				return;
			}
			// look for our interrupt endpoint
			for (int i = 0; i < intf.getEndpointCount(); i++) {
				log2comproto("Endpoint: "+intf.toString(),2);
				UsbEndpoint ep = intf.getEndpoint(i);
				if (ep.getType() == UsbConstants.USB_ENDPOINT_XFER_INT) {
					if (ep.getDirection() == UsbConstants.USB_DIR_IN) {
						epIn = ep;
					}
				}

			}
			if ( epIn == null) {
				throw new IllegalArgumentException("not all endpoints found.");
			} else log2comproto("Endpoint found.",1);
			if (epIn.getType() != UsbConstants.USB_ENDPOINT_XFER_INT) {
				log2comproto("ERROR: endpoint is not interrupt type!",0);
			}


		}

		EndpointOut = epOut;
		EndpointIn = epIn; 
		Device = device;
		Vid=device.getVendorId();
		Pid=device.getProductId();
		Serial_id="00_0000000000";
		Manufacturer="UKN";
		Product="unknown logger";

		/* Get Vendor name, Product name and Serial-Nr of Logger*/
		String ana=USBTools.getUsbInfoViaShell();
		String[] sep=ana.split("\n");
		int i;
		String a=String.format("%04x", Vid);
		String b=String.format("%04x", Pid);

		for(i=0;i<sep.length; i++) {
			String[] sep2=sep[i].split(",");
			if(sep2.length>2) {
				String c=sep2[0].replace("\"","");
				String d=sep2[1].replace("\"","");
				if(a.equalsIgnoreCase(c) && b.equalsIgnoreCase(d)) {
					if(sep2.length>2) {
						Serial_id=sep2[2].replace("\"","");
					} 
					if(sep2.length>3) {
						Manufacturer=sep2[3].replace("\"","");
					} 
					if(sep2.length>4) {
						Product=sep2[4].replace("\"","");
					} 
				}
			}
		} 

		UsbDeviceConnection connection = UsbManager.openDevice(device);
		if (connection != null && connection.claimInterface(intf, true)) {
			log2comproto("open SUCCESS: "+connection.toString(),1);
			isconnected=true;

			Connection = connection;

			if(Vid==6465 && (Pid==2010 || Pid==32801)) {
				/*TODO: HID specific:*/

				//	Connection.controlTransfer(UsbConstants.USB_DIR_OUT|0x21, 0x9,0x0, 0x1, null, 0, WAIT_WRITE);
				loggertype=LTYP_TH;
				packet_size=8;
				memory_size=0x10000;	
				if(Manufacturer.equals("UKN") && Product.equals("unknown logger")) {
					Manufacturer="FreeTec";
					Product="USB data logger";
				}

			} else if(Vid==4292 && Pid==60000) {  // VDL-141 ...

		        setConfigSingle(CP210x.SILABSER_IFC_ENABLE_REQUEST_CODE, CP210x.UART_ENABLE);
		        setConfigSingle(CP210x.SILABSER_SET_MHS_REQUEST_CODE, CP210x.MCR_ALL | CP210x.CONTROL_WRITE_DTR | CP210x.CONTROL_WRITE_RTS);
		        setConfigSingle(CP210x.SILABSER_SET_BAUDDIV_REQUEST_CODE, CP210x.BAUD_RATE_GEN_FREQ / CP210x.DEFAULT_BAUD_RATE);            
//		            setParameters(DEFAULT_BAUD_RATE, DEFAULT_DATA_BITS, DEFAULT_STOP_BITS, DEFAULT_PARITY);
		  
				protocol=PROTO_CP210X;
				packet_size=64;
				memory_size=0x10000;
				loggertype=LTYP_TH;
				calibration_Avalue=(float)0.1;
				calibration_Bvalue=0;
				calibration_A2value=(float)0.1;
				calibration_B2value=0;
			} else if(Vid==4292 && Pid==60001) {
				/*Das scheint nicht noetig zu sein ... */
				reset();
				lock_device();       /* USB lock device*/
				/* USB unlock device (hier weglassen, da spaeter?)*/
				Connection.controlTransfer(0x40, 2, 2, 0, null, 0, WAIT_WRITE);
				/* Hier jetzt unterscheiden, welcher Typ Datenlogger angeschlossen ist.*/
				if(Product.equals("Weather Datalogger")) {
					protocol=PROTO_VOLTCRAFT_WEATHER;
					loggertype=LTYP_THP;
					memory_size=0x10000;
					packet_size=64;
				} else {
					protocol=PROTO_VOLTCRAFT_NEW;
					packet_size=64;
					memory_size=0xfe00;
				}
				calibration_Avalue=(float)0.1;
				calibration_Bvalue=0;
				calibration_A2value=(float)0.1;
				calibration_B2value=0;
			} else if(Vid==4292 && Pid==2) {
				protocol=PROTO_ELV;
				if(Manufacturer.equals("UKN") && Product.equals("unknown logger")) {
					Manufacturer="USB500";
					Product="data logger";
				}
				/* TODO */
				/* Hier muss dann eigentlich sofort get_config ausgeführt
				 * werden, sonst inst die Calibration inkonsistent, sollte man gleich ein 
				 * read data machen....*/
				config.block_type=0;  /*Set as unknown*/
				calibration_Avalue=(float)1;
				calibration_Bvalue=0;
				calibration_A2value=(float)0.5;
				calibration_B2value=0;
			} else {
				/*The old types do not need configuration */
				protocol=PROTO_VOLTCRAFT;
				calibration_Avalue=(float)0.1;
				calibration_Bvalue=0;
				calibration_A2value=(float)0.1;
				calibration_B2value=0;
				packet_size=64;
				memory_size=0xfe00;
				if(Manufacturer.equals("UKN") && Product.equals("unknown logger")) {
					Manufacturer="generic";
					Product="data logger";
				}
			}
			mActivity.displaystatus(Manufacturer+" "+Product+" (#"+Serial_id+") connected.",0);
		} else {
			mActivity.displaymessage("ERROR: Device open failed!",5);
			Connection = null;
		}    
	}

	public void close() {
		log2comproto("### close",1);
		if(Connection!=null) {
			Connection.close();
			Connection=null;
		}
		simulator.sim_configbuf=null;
		isconnected=false;
	}
	public int receive(byte[] cbuf)  {
		if(do_simulate) {
			return simulator.send(cbuf);
		} else {
			int ret=Error.ERR;
			if(Connection != null) {
				ret=Connection.bulkTransfer(EndpointIn, cbuf, cbuf.length,WAIT_READ);
			} else log2comproto("receive: connection problem",0);
			log2comproto("received: ret="+ret+" "+bytearray2string(cbuf),1);

			/*Mache es etwas langsamer um Übertragungsfehler bei altem Logger
			 * zu vermeiden. Es wurden ab dem 17. Block daten schonmal ganz andere
			 * seltsame Speicherbereiche empfangen...*/
			if(protocol==PROTO_VOLTCRAFT) {
				try {
					Thread.sleep(50);
				} catch (InterruptedException e1) {
					log2comproto("sleep was interrupted: "+e1.getMessage(),1);
				}
			}

			return(ret);
		}
	}
	public int receive(byte[] cbuf,int n) {
		if(do_simulate) {
			return simulator.send(cbuf,n);
		} else {
			int ret=Error.ERR;
			if(Connection != null) {
				ret=Connection.bulkTransfer(EndpointIn, cbuf,n,WAIT_READ);
			} else log2comproto("receive: connection problem",0);
			log2comproto("received: ret="+ret+" "+bytearray2string(cbuf),1);
			return(ret);
		}
	}
	public int receive(byte[] cbuf,int n,int timeout) {
		if(do_simulate) {
			return simulator.send(cbuf,n);
		} else {
			int ret=Error.ERR;
			if(Connection != null) {
				ret=Connection.bulkTransfer(EndpointIn, cbuf,n,timeout);
			} else log2comproto("receive: connection problem",0);
			log2comproto("received: ret="+ret+" "+bytearray2string(cbuf),1);
			return(ret);
		}
	}
	public int send(byte[] cbuf) {
		log2comproto("send "+bytearray2string(cbuf),1); 
		if(do_simulate) {
			return simulator.receive(cbuf);
		} else {
			int ret=Error.ERR;
			if(Connection != null) {
				ret=Connection.bulkTransfer(EndpointOut, cbuf, cbuf.length,WAIT_WRITE);
			} else log2comproto("send: connection problem",0);
			return(ret);
		}
	}
    private static final int REQTYPE_HOST_TO_DEVICE = 0x41;

	private int setConfigSingle(int request, int value) {
	   return Connection.controlTransfer(REQTYPE_HOST_TO_DEVICE, request, value, 
	                0, null, 0, WAIT_WRITE);
	}

	
	
	public byte [] receiveHID() {
		if(do_simulate) {
			return simulator.receiveHID();
		} else {

			byte [] cbuf;
			ByteBuffer buffer = ByteBuffer.allocate(8);
			if(Connection != null) {
				buffer.put(0,(byte)0x20);
				UsbRequest request = new UsbRequest();
				request.initialize(Connection, EndpointIn);
				request.queue(buffer, 8);

				if (Connection.requestWait() == request) {
					try {
						Thread.sleep(1);
					} catch (InterruptedException e) {
					}
				} else {
					log2comproto("requestWait failed, exiting",0);
				}

				// ret=Connection.bulkTransfer(EndpointIn, cbuf,n,WAIT_READ);
			} else log2comproto("receive: connection problem",0);
			cbuf=buffer.array();
			log2comproto("received: "+bytearray2string(cbuf),2);
			return(cbuf);
		}
	}

	public void unlock_device(){
		log2comproto("unlock",1);
		if(Connection!=null) Connection.controlTransfer(0x40, 2, 2, 0, null, 0, WAIT_WRITE);
	}
	public void lock_device(){
		log2comproto("lock",1);
		if(Connection!=null) Connection.controlTransfer(0x40, 2, 4, 0, null, 0, WAIT_WRITE);
	}
	public void reset(){
		log2comproto("reset",1);
		if(Connection!=null) Connection.controlTransfer(0x40, 0, 0xff, 0xff, null, 0, Logger.WAIT_WRITE);
	}


	/*Send a command to the device
	 * */

	public void sendCommand(byte a,byte b,byte c) {
		log2comproto("sendCommand " + a+":"+b+":"+c,1);
		byte[] message = new byte[3];
		message[0] = a;
		message[1] = b;
		message[2] = c; 
		send(message);
	}
	public void sendHIDCommand(int control,int bank, int adr,int data) {
		if(do_simulate) {
			simulator.sendHIDCommand(control,bank,adr,data);
		} else {
			synchronized (this) {
				if(Connection!=null) {
					byte[] message = new byte[8];
					message[0] = (byte)control; 
					message[1] = (byte)(bank & 0xff);
					message[2] = (byte)(adr & 0xff);
					message[3] = (byte)0x20;
					message[4] = (byte)control;
					message[5] = (byte)data;
					message[6] = (byte)0;
					message[7] = (byte)0x20;
					// Send command via a control request on endpoint zero
					Connection.controlTransfer(0x21, 0x9, 0x200, 0, message, message.length, 0);
					log2comproto("SendHIDCommand: "+bytearray2string(message),2);
				}
			}
		}
	}
	public void sendHIDData(byte[] message) {
		if(do_simulate) {
			simulator.sendHIDData(message);
		} else {

			synchronized (this) {
				if(Connection!=null) {
					// Send command via a control request on endpoint zero
					Connection.controlTransfer(0x21, 0x9, 0x200, 0, message, message.length, 0);
					log2comproto("SendHIDData: "+bytearray2string(message),2);
				}
			}
		}
	}

	/*Read a 1 byte Acknoledge message, all protocols*/

	public int readack() {
		byte ack[]=new byte[1];
		ack[0]=0x00;
		int ret=0;
		int trys=0;
		/* Sometimes restarting the logger takes several seconds before it sends the ACK*/
		while((ack[0]&0xff)!=0xff && trys<5) {
			ret=receive(ack);
			log2comproto("Wait for ACK, response: ret="+ret+" ack="+ack[0],1);
			trys++;
		}
		if(ret==1) {
			if((ack[0]&0xff)==0xff) return 1;
			else return 0;	
		}
		return Error.ERR;  /*Something is wrong: timeout*/
	}



	/*Read a 3 bytes response or status message*/

	public int readresponse(byte a,byte b,byte c) {
		log2comproto("Readresponse",1);

		int ret;
		int trys=0;
		int size;
		byte ack[]=new byte[3];
		ack[0]=(byte) 0;   		
		if(protocol==PROTO_VOLTCRAFT_WEATHER) {
			while((ack[0]&0xff)!=0x0f && (ack[0]&0xff)!=0xfe) {
				ret=receive(ack);
				log2comproto(" ret="+ret+" ack="+ack[0],1);
				if(ret==1) {  /* could be an acknoledge */
					if((ack[0]&0xff)==0xff) return -3; /*handshake packet 0xFF, after FLASH write */
					else return 0;
				}
				if (trys > 10) {
					log2comproto(" read_abc: Stopping now.",0);
					return(Error.ERR);
				}
				trys++;
			}
		} else {
			while(ack[0]!=0x02) {

				ret=receive(ack);
				log2comproto(" ret="+ret+" ack="+ack[0],1);
				if((ret==3 && ack[0] != 0x2)) {
					log2comproto(" read_abc: try = "+trys+" byte = "+ack[0],1);
					sendCommand(a,b,c);
				} else if(ret==1) {  /* could be an acknoledge */
					if((ack[0]&0xff)==0xff) return -3; /*handshake packet 0xFF, after FLASH write */
					else return 0;
				}
				if (trys > 10) {
					log2comproto(" read_abc: Stopping now.",0);
					return(Error.ERR);
				}
				trys++;
			}
		}
		log2comproto(" ack="+ack[0]+" "+ack[1]+" "+ack[2],1);
		if(protocol==PROTO_VOLTCRAFT_WEATHER && (ack[0]&0xff)==0xf) size = (((ack[1]&0xff) << 8) | (ack[2]&0xff));
		else if(protocol==PROTO_VOLTCRAFT_WEATHER && (ack[0]&0xff)==0xfe && ack[1]==1) size = 0;
		else size = (((ack[2]&0xff) << 8) | (ack[1]&0xff));
		log2comproto(" size="+size,1);
		return(size);
	}

	public int sendCommandResponse(int control) {
		byte a,b,c;
		if(control==-1) {  /*Read config*/
			if(protocol==PROTO_VOLTCRAFT_WEATHER) {a=(byte)0xf;b=(byte)0;c=(byte)0;}
			else {a=(byte)0x0;b=(byte)16;c=(byte)1;}
		} else if(control==-2) {  /*Write config*/
			if(protocol==PROTO_VOLTCRAFT_WEATHER) {a=(byte)0xe;b=(byte)0x40;c=(byte)0;}
			else {a=(byte)0x1;b=(byte)0x40;c=(byte)0;}
		} else {            /*Read data bank*/
			if(protocol==PROTO_VOLTCRAFT_WEATHER) {a=(byte)0xf;b=(byte)(1+control);c=(byte)0;}
			else if(protocol==PROTO_ELV) {a=(byte)0x3;b=(byte)(0xff);c=(byte)0xff;}
			else {a=(byte)0x0;b=(byte)(control&0xf);c=(byte)0x40;}
		}
		sendCommand(a,b,c);
		return readresponse(a,b,c);
	}
	/*Read configuration from device
	 * */
	public int get_config() {
		log2comproto("### getconfig",1);
		if(protocol==PROTO_ELV) return Lascar.get_config(this);
		else if(protocol==PROTO_HID) return Freetec.get_config(this);
		else if(protocol==PROTO_CP210X) return CP210x.get_config(this);
		else return Voltcraft.get_config(this);
	}
	public int get_data() {
		log2comproto("### getdata",1);
		data.clear();
		data.set_calibration(calibration_Avalue,calibration_Bvalue,calibration_A2value,calibration_B2value);
		int ret;
		if(config.num_data_rec==0) return(Error.OK);
		if(protocol==PROTO_ELV) ret=Lascar.get_data(this);
		else if(protocol==PROTO_HID) ret=Freetec.get_data(this);
		else if(protocol==PROTO_CP210X) ret=CP210x.get_data(this);
		else ret=Voltcraft.get_data(this);
		log2comproto("num_data ="+ data.anzdata,1);
		if(protocol!=PROTO_HID) data.anzdata=config.num_data_rec;  /*Hm....*/
		isreaddata=true;
		return ret;
	}
	public int chk_conf() {
		log2comproto("chkconf",1);
		int ret=config.chk_name();
		if(ret!=Error.OK) return ret; 
		if(protocol==Logger.PROTO_ELV) return Lascar.chk_conf(config, loggertype, do_repair);
		else if(protocol==PROTO_HID) return Freetec.chk_conf(config, loggertype);
		else if(protocol==PROTO_CP210X) return CP210x.chk_conf(config, loggertype,memory_size);
		else return Voltcraft.chk_conf(config, loggertype,memory_size);
	}
	public byte[] build_conf() {
		log2comproto("buildconf",1);
		if(protocol==Logger.PROTO_ELV) return Lascar.build_conf(this);
		else if(protocol==PROTO_HID) return Freetec.build_conf(this);
		else if(protocol==PROTO_CP210X) return CP210x.build_conf(this);
		else return Voltcraft.build_conf(config, loggertype, protocol);
	}

	public String get_status() {
		log2comproto("getstatus",1);
		if(protocol==Logger.PROTO_ELV) return Lascar.get_status(this);
		else if(protocol==PROTO_HID) return Freetec.get_status(this);
		else if(protocol==PROTO_CP210X) return CP210x.get_status(this);
		else return Voltcraft.get_status(this);
	}

	public int send_config(byte[] confmessagebuf) {
		log2comproto("### sendconfig "+bytearray2string(confmessagebuf),1);
		if(isconnected) {
			if(protocol==PROTO_ELV ) {
				Lascar.stop_logger(this);
				return Lascar.write_configblock(confmessagebuf,this);
			} else if(protocol==PROTO_HID) {
				return Freetec.write_configblock(confmessagebuf,this);
			} else if(protocol==PROTO_CP210X) {
				return CP210x.write_configblock(confmessagebuf,this);
			} else {
				if(protocol==PROTO_VOLTCRAFT_WEATHER ||protocol==PROTO_VOLTCRAFT_NEW || protocol==PROTO_ELV ) unlock_device();
				if(protocol==PROTO_VOLTCRAFT_WEATHER) sendCommand((byte)0xe,(byte)(packet_size&0xff),(byte)((packet_size>>8)&0xff));
				else sendCommand((byte)0x1,(byte)(packet_size&0xff),(byte)((packet_size>>8)&0xff));
				send(confmessagebuf);
				if(readack()!=1) return(Error.ERR);
				if(protocol==PROTO_VOLTCRAFT_WEATHER ||protocol==PROTO_VOLTCRAFT_NEW || protocol==PROTO_ELV ) lock_device();
				return(Error.OK);
			}
		} else {
			log2comproto("sendconfig: connection problem",0);
			return(Error.ERR);
		}
	}
	public String getStatistics() {	
		String infostring="["+String.format("%04d-%02d-%02d %02d:%02d:%02d",config.time_year,(int)(config.time_mon&0xff),
				(int)(config.time_mday&0xff),(int)(config.time_hour&0xff),(int)(config.time_min&0xff),
				(int)(config.time_sec&0xff))+"] "+config.num_data_rec+" points @ "+config.interval+" sec.";

		String loggerinfostring=Manufacturer+"-"+Product+"/"+loggertype+"("+typestring()+")"+" #"+Serial_id+" blocktype="+config.block_type
				+" Session: "+config.getname();
		String calibrationinfostring="A="+data.calibration_Avalue+" B="+data.calibration_Bvalue+" "+
				"A2="+data.calibration_A2value+" B2="+data.calibration_B2value+" "+
				"M="+config.calibration_Mvalue+" C="+config.calibration_Cvalue;
		String settingsinfo="";
		String statistics="";
		String eventstat="";
		if(loggertype==Logger.LTYP_THP) 
			settingsinfo+=String.format(Locale.US,"# Pressure limits: [%.2f:%.2f] hPa<br/>",config.thresh_press_low,config.thresh_press_high);
		if(loggertype==Logger.LTYP_THP || loggertype==Logger.LTYP_TH) 
			settingsinfo+=String.format(Locale.US,"# Humidity limits: [%.1f:%.1f] %%<br/>",config.thresh_rh_low,config.thresh_rh_high);
		if(loggertype==Logger.LTYP_THP || loggertype==Logger.LTYP_TH || loggertype==Logger.LTYP_TEMP) 
			settingsinfo+=String.format(Locale.US,"# Temperature limits: [%.1f:%.1f] %s<br/>",config.thresh_temp_low,config.thresh_temp_high,
					((config.temp_is_fahrenheit&1)>0?"°F":"°C"));
		if(loggertype==Logger.LTYP_VOLT || loggertype==Logger.LTYP_CURR)
			settingsinfo+=String.format(Locale.US,"# Limits: [%f:%f] %s<br/>",config.thresh_temp_low,config.thresh_temp_high,config.getunit());

		if(config.start==1) settingsinfo+=String.format(Locale.US,"# Start delay=%d sec<br/>",config.time_delay);
		settingsinfo+="Rawinput="+config.rawinputreading+" flagbits="+config.flag_bits+"<br/>";

		/*Datenstatistik erstellen
		 * */
		Calendar cal = Calendar.getInstance();
		cal.set(config.time_year, (int)(config.time_mon&0xff)-1, (int)(config.time_mday&0xff), (int)(config.time_hour&0xff), (int)(config.time_min&0xff), (int)(config.time_sec&0xff));
		long ss=cal.getTimeInMillis()/1000;
		statistics+="";
		{
			int duration=data.anzdata*config.interval;
			int d=duration;
			int s=d%60;
			d=(d-s)/60;
			int m=d%60;
			d=(d-m)/60;
			int h=d%24;
			d=(d-h)/24;

			if(duration>0) statistics+="Time range: ["+ss+":"+(ss+duration)+"] "+String.format("%dd%02d:%02d:%02d",d,h,m,s)+"<br/>";
		}

		float tmin=9999,tmax=-9999,t;
		for(int i=0;i<data.anzdata;i++) {
			t=data.get_temp(i);
			tmin=Math.min(tmin,t);
			tmax=Math.max(tmax,t);
		}
		if(loggertype==Logger.LTYP_THP || loggertype==Logger.LTYP_TH || loggertype==Logger.LTYP_TEMP) 
			statistics=statistics+String.format(Locale.US,"Temperature range: [%.1f:%.1f] %s<br/>",
					tmin,tmax,((config.temp_is_fahrenheit&1)>0?"°F":"°C"));
		else if(loggertype==Logger.LTYP_VOLT || loggertype==Logger.LTYP_CURR) 
			statistics=statistics+String.format(Locale.US,"Range: [%f:%f] %s<br/>",tmin,tmax,config.getunit());;
			if(loggertype==Logger.LTYP_THP || loggertype==Logger.LTYP_TH) {
				float rhmin=9999,rhmax=-9999,r;
				for(int i=0;i<data.anzdata;i++) {
					r=data.get_rh(i);
					rhmin=Math.min(rhmin,r);
					rhmax=Math.max(rhmax,r);
				} 
				statistics=statistics+String.format(Locale.US,"Humidity range:    [%.1f:%.1f] %%<br/>",
						rhmin,rhmax);	
				/* Events zusammenfassen   */

				if(loggertype==Logger.LTYP_THP || loggertype==Logger.LTYP_TH) {
					if(data.eventliste!=null && data.eventliste.size()>0) {
						eventstat+="<p/> Detected "+data.eventliste.size()+" Events:<br/>";
						Lueftungsevent a;
						for(int i=0;i<data.eventliste.size();i++) {
							a=data.eventliste.get(i);
							eventstat+="# "+i+": "+Physics.timestamp2datetime_short(ss+config.interval*a.idx)+
									" for "+Physics.gerundet(config.interval*(a.idx2-a.idx)/60,1)+" minutes.<br/>";
						}
						eventstat+="<p/>";
					}
				}

			}
			if(loggertype==Logger.LTYP_THP) {
				float pmin=9999,pmax=-9999;
				for(int i=0;i<data.anzdata;i++) {
					pmin=Math.min(pmin,data.p[i]);
					pmax=Math.max(pmax,data.p[i]);
				} 
				statistics=statistics+String.format(Locale.US,"# Pressure range:    [%.2f:%.2f] hPa<br/>",
						Voltcraft.raw2p((short)pmin),Voltcraft.raw2p((short)pmax));

			}
			if(eventstat.length()>0) statistics+=eventstat;

			return "<h2>Logger:</h2> "+loggerinfostring+
					"<h2>Calibration:</h2>"+calibrationinfostring+
					"<h2>Settings:</h2>"+settingsinfo+
					"<h2>Data Statistics:</h2>"+statistics+"<p/>"+infostring;

	}
	public String typestring() {
		if(loggertype==LTYP_TEMP) return "Temperature logger";
		else if(loggertype==LTYP_TH) return "Temperature+Humidity logger";
		else if(loggertype==LTYP_THP) return "Temperature+Humidity+Pressure logger";
		else if(loggertype==LTYP_VOLT) return "Voltage logger";
		else if(loggertype==LTYP_CURR) return "Current logger";
		else if(loggertype==LTYP_GAS) return "Gas concentration logger";
		else if(loggertype==LTYP_SOUND) return "Sound level logger";
		else return "logger";
	}
	public void updateProgress(int n) {
		if(be_quiet==false && mActivity!=null) mActivity.updateProgress(n);
	}
	//public void displayMessage(String s,int n) {
	//	if(be_quiet==false && mActivity!=null) mActivity.displaymessage(s,n);
	//}	
	public void updateMessage(String s,int n) {
		if(be_quiet==false && mActivity!=null) mActivity.updateMessage(s,n);
	}	
	public void updateStatus(String s,int n) {
		if(be_quiet==false && mActivity!=null) mActivity.updateStatus(s,n);
	}	
	public void updateStatus(int n) {
		if(be_quiet==false && mActivity!=null) mActivity.updateStatus(n);
	}
}
